{ This demo program shows how to use the file I/O functions implemented
  in ChroMathsome - Statistics.
  Also added are demonstrations on how to retrieve data points by using a TStringGrid and
  the use of a statistical matrix (StatMatrix).

  The reader may do with this code whatever he/she wants.
  Author: John Merrit (john.merrit@net4all.be)
  Release date: May '98

  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  ChroMathsome 1996-98 John Merrit. All Rights Reserved

  Brought to you by the letter "S".             }

unit IFace;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, StdCtrls, ComCtrls, Statistc;

type
  TForm1 = class(TForm)
    MainMenu1: TMainMenu;
    Close1: TMenuItem;
    Fileio1: TMenuItem;
    SaveStatVectortoFile1: TMenuItem;
    LoadStatVectorFromFile1: TMenuItem;
    OpenDialog1: TOpenDialog;
    SaveDialog1: TSaveDialog;
    Memo1: TRichEdit;
    StatMatrix1: TMenuItem;
    UseStatVectors1: TMenuItem;
    EnterStatVector1: TMenuItem;
    procedure Close1Click(Sender: TObject);
    procedure SaveStatVectortoFile1Click(Sender: TObject);
    procedure LoadStatVectorFromFile1Click(Sender: TObject);
    procedure UseStatVectors1Click(Sender: TObject);
    procedure EnterStatVector1Click(Sender: TObject);
    
  private
    { Private declarations }
    procedure Write_StatVector(V: StatVector);
    procedure Write_StatMatrix(M: StatMatrix);
    procedure GetStatVectorFromGrid(var V: StatVector; VectorIndex: integer);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

uses DPointD; // Data point grid.

procedure TForm1.Close1Click(Sender: TObject);
begin
     Close;
end;

{ Write a statvector to the screen, creating a nice readable look. }
procedure TForm1.Write_StatVector(V: StatVector);
var a:integer;
    s,ss,st:string;
begin
    Memo1.Enabled:= true;
    st:= '';
    ss:= '[ ';
    for a:= 1 to V.Dimension do
    begin
    if abs(V.Vector^[a]) < 1E-18 then
    begin
    V.Vector^[a]:= 0;
    s:= IntToStr(round(V.Vector^[a]));
    end
    else
    s:= FloatToStrF(V.Vector^[a], ffexponent,5,2);
    st:= st + s + ' ';
    end;
    ss:= ss + st + ']';
    Memo1.Lines.add(ss);
end;

{ Write a statmatrix to the screen, creating a nice readable look. }
procedure TForm1.Write_StatMatrix(M: StatMatrix);
var a,b:integer;
    matt,mats,mattt:string;
begin
    memo1.Enabled:= true;
    mattt:= '';
    for a:= 1  to M.Row do
    begin
    mats:= '[ ';
    for b:= 1 to M.Col do
    begin
    if abs(M.Matrix^[a]^[b]) < 1E-18 then
    begin
    M.Matrix^[a]^[b]:= 0;
    matt:= '0';
    end
    else
    matt:= FloatToStrF(M.Matrix^[a]^[b], ffexponent,5,2);
    mattt:= mattt + matt + ' ';
    end;
    mats:= mats + mattt + ']';
    mattt:= '';
    memo1.Lines.add(mats);
    end;
end;

procedure TForm1.GetStatVectorFromGrid(var V: StatVector; VectorIndex: integer);
var i, j: integer;
    x: extended;
    Dimension: integer;
begin
     Dimension:= DataPointDialog.DataPointGrid.ColCount - 1;
     j:= DataPointDialog.DataPointGrid.RowCount - 1;
     if ((VectorIndex <= 0) or (VectorIndex > j)) then
     begin
       Memo1.Lines.Add('Not a valid statvector index given');
       abort;
     end
     else
     begin
       RedimensionStatVector(V, Dimension);
       for i:= 1 to Dimension do
       begin
         x:= StrToFloat(DataPointDialog.DataPointGrid.Cells[i, VectorIndex]);
         V.Vector^[i]:= x;
       end;
     end;
end;

procedure TForm1.SaveStatVectortoFile1Click(Sender: TObject);
var V: StatVector;
begin
     if SaveDialog1.Execute then
     begin
           InitStatVector(V, 5);
           try
           V.Vector^[1]:= 5;
           V.Vector^[2]:= 2;
           V.Vector^[3]:= 4;
           V.Vector^[4]:= 3;
           V.Vector^[5]:= 7;
           SaveStatVectorToFile(V, SaveDialog1.FileName);
           Memo1.Lines.Add('StatVector saved in file ' + SaveDialog1.FileName);
           finally
           DestroyStatVector(V);
           end;
     end
end;

procedure TForm1.LoadStatVectorFromFile1Click(Sender: TObject);
var V: StatVector;
begin
     if OpenDialog1.Execute then
     begin
           InitStatVector(V, 1);
           try
           LoadStatVectorFromFile(V, OpenDialog1.FileName);
           Memo1.Lines.Add('Loaded StatVector: ');
           Write_StatVector(V);
           Memo1.Lines.Add('Sum of elements of statvector: ' + FloatToStr(Sum(V)));
           finally
           DestroyStatVector(V);
           end;
     end
end;

{ Get statvectors from a statmatrix ... }
procedure TForm1.UseStatVectors1Click(Sender: TObject);
var S: StatMatrix;
    V1, V2: StatVector;
    x: extended;
    i, j: integer;
begin
     InitStatMatrix(S, 20, 20);
     InitStatVector(V1, 1); // 1 is dummy value ...
     InitStatVector(V2, 1); // 1 ...

     randomize;
     try
     // Initialize matrix S with random numbers
     for i:= 1 to S.Row do
     for j:= 1 to S.Col do
     S.Matrix^[i]^[j]:= random(100) + 1;

     // Write statmatrix to screen for testing.
     Memo1.Lines.Add('StatMatrix S:');
     Write_StatMatrix(S);

     // Calculate mean of every column (columns represent statvectors!).
     // Save means in statvector V2
     for i:= 1 to S.Col do
     begin
       StatMatrixColToStatVector(V1, S, i);
       x:= Mean(V1);
       if i = 1 then V2.Vector^[1]:= x // V2 is initialized to be a 1-dimensional vector.
       else
       StatVectorAddElement(V2, x);
     end;

     // Write means in memo.
     for i:= 1 to V2.Dimension do
     Memo1.Lines.Add('Mean of statvector ' + IntToStr(i) + ': ' + FloatToStr(V2.Vector^[i]));

     finally
     DestroyStatMatrix(S);
     DestroyStatVector(V1);
     DestroyStatVector(V2);
     end;
end;

{ Use Data Point Grid to retrieve a statvector. }
procedure TForm1.EnterStatVector1Click(Sender: TObject);
var V: StatVector;
begin
     DataPointDialog.DataPointGrid.RowCount:= 3;
     DataPointDialog.ShowModal;
     InitStatVector(V, 1); // 1 is dummy value ...
     try
     Memo1.Lines.Add('StatVector 1:');
     GetStatVectorFromGrid(V, 1); // ... updates dimension of V if needed.
     Write_StatVector(V);

     Memo1.Lines.Add('StatVector 2:');
     GetStatVectorFromGrid(V, 2); // ... updates dimension of V if needed.
     Write_StatVector(V);
     Memo1.Lines.Add('Sum of elements of statvector 2: ' + FloatToStr(Sum(V)));

     finally
     DestroyStatVector(V);
     end;
end;

end.
