{ Demonstration program for ChroMathsome - Statistics.

  This demo program shows how we can expand Chromathsome - Statistics.
  As an example, we'll try to fit data points (Xi, Yi) into a curve that is
  a linear combination of functions:
  f(x) = a1 + a2*x + a3*sin(x) + a4*exp(x).

  Instead of just telling you how to do it, we've built the solution up, step by step.
  Several help functions are added needed to find the solution.

  The same analytical tools and techniques can be applied for many other regression functions.

  This demo, again, shows how easy it is to manipulate data sets in statmatrices and statvectors.


  The reader may do with this code whatever he/she wants.
  Author: John Merrit (john.merrit@net4all.be)
  Release date: May '98

  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  ChroMathsome - Statistics 1996-98 John Merrit. All Rights Reserved

  Brought to you by the letter "E".             }
unit expan;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Menus, Statistc, FlMath;

type
  TForm1 = class(TForm)
    Memo1: TMemo;
    MainMenu1: TMainMenu;
    Close1: TMenuItem;
    CurveFit1: TMenuItem;
    fxa1a2xa3sinxa4expx1: TMenuItem;
    procedure Close1Click(Sender: TObject);
    procedure fxa1a2xa3sinxa4expx1Click(Sender: TObject);
  private
    { Private declarations }
    NumOfSwaps: integer; // Sign of StatMatrix determinant.
    procedure StatMatrixFindPivotElement(V: StatVector; var PivotValue: extended; var Row: integer; StartIndex: integer);
    procedure StatMatrixGaussNormalForm(var M: StatMatrix);
    function StatMatrixDeterminant(M: StatMatrix): extended;

    procedure Write_StatVector(V: StatVector);
    procedure Write_StatMatrix(M: StatMatrix);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

{ Write a statvector to the screen, creating a nice readable look. }
procedure TForm1.Write_StatVector(V: StatVector);
var a:integer;
    s,ss,st:string;
begin
    Memo1.Enabled:= true;
    st:= '';
    ss:= '[ ';
    for a:= 1 to V.Dimension do
    begin
    if abs(V.Vector^[a]) < 1E-18 then
    begin
    V.Vector^[a]:= 0;
    s:= IntToStr(round(V.Vector^[a]));
    end
    else
    s:= FloatToStrF(V.Vector^[a], ffexponent,5,2);
    st:= st + s + ' ';
    end;
    ss:= ss + st + ']';
    Memo1.Lines.add(ss);
end;

{ Write a statmatrix to the screen, creating a nice readable look. }
procedure TForm1.Write_StatMatrix(M: StatMatrix);
var a,b:integer;
    matt,mats,mattt:string;
begin
    memo1.Enabled:= true;
    mattt:= '';
    for a:= 1  to M.Row do
    begin
    mats:= '[ ';
    for b:= 1 to M.Col do
    begin
    if abs(M.Matrix^[a]^[b]) < 1E-18 then
    begin
    M.Matrix^[a]^[b]:= 0;
    matt:= '0';
    end
    else
    matt:= FloatToStrF(M.Matrix^[a]^[b], ffexponent,5,2);
    mattt:= mattt + matt + ' ';
    end;
    mats:= mats + mattt + ']';
    mattt:= '';
    memo1.Lines.add(mats);
    end;
end;


procedure TForm1.Close1Click(Sender: TObject);
begin
     Close;
end;

{ Pivoting -> Gauss. }
procedure TForm1.StatMatrixFindPivotElement(V: StatVector; var PivotValue: extended; var Row: integer; StartIndex: integer);
var i, j: integer;
    x: extended;
begin
     x:= abs(V.Vector^[StartIndex]);
     j:= StartIndex;
     for i:= StartIndex + 1 to V.Dimension do
     begin
     if abs(V.Vector^[i]) > x then
     begin
     j:= i;
     x:= abs(V.Vector^[i]);
     end;
     end;
     PivotValue:= V.Vector^[j];
     Row:= j;
end;

{ Gauss elimination. Simplified for a square matrix!!! (original code: ChroMathsome - Linear Algebra) }
procedure TForm1.StatMatrixGaussNormalForm(var M: StatMatrix);
var Row, i, j, n, o:integer;
    V: StatVector;
    PivotValue, PivotCoefficient: extended;
begin
      if M.Row > 1 then
      begin
      InitStatVector(V, M.Row);
      try
      NumOfSwaps:= 0; { # of swapped rows -> sign of determinant }
      if (M.Row > M.Col) then j:= M.Col else j:= M.Row;
      for i:= 1 to j-1 do
      begin
      StatMatrixColToStatVector(V, M, i);
      StatMatrixFindPivotElement(V, PivotValue, Row, i);
      if Row <> i then
      begin
      NumOfSwaps:= NumOfSwaps + 1;
      StatMatrixSwapRows(M, i, Row);
      end;
      if PivotValue <> 0 then
      begin
        for n:= i+1 to M.Row do
          begin
             PivotCoefficient:= M.Matrix^[n]^[i]/PivotValue;
             for o:= i to M.Col do
             M.Matrix^[n]^[o]:= M.Matrix^[n]^[o] - PivotCoefficient*M.Matrix^[i]^[o];
          end;
      end;
      end;
     finally
     DestroyStatVector(V);
     end;
     end;
end;

{ Determinant of square matrix.  }
function TForm1.StatMatrixDeterminant(M: StatMatrix): extended;
var x: extended;
    i: integer;
begin
     x:= 1;
     StatMatrixGaussNormalForm(M);
     for i:= 1 to M.Row do
     x:= x*M.Matrix^[i]^[i];
     if odd(NumOfSwaps) then StatMatrixDeterminant:= -x
     else StatMatrixDeterminant:= x;
end;

{ Fit data points X = [ 0.1 0.4 0.5 0.7 0.7 0.9 ] and
  Y = [ 0.61 0.92 0.99 1.52 1.47 2.03 ] into a curve f(x) = a1 + a2*x + a3*sin(x) + a4*exp(x).
  Note: we assume the reader is familiar with the method of least squares.                     }
procedure TForm1.fxa1a2xa3sinxa4expx1Click(Sender: TObject);
var M, M1: StatMatrix;
    X, X1, Y, Y1: StatVector;
    det, cf: extended;
    i: integer;
begin
     InitStatMatrix(M, 1, 1);
     InitStatMatrix(M1, 1, 1);
     InitStatVector(X, 1);
     InitStatVector(X1, 1);
     InitStatVector(Y, 1);
     InitStatVector(Y1, 1);

     try
     { Populate statvectors X and Y. }
     RedimensionStatVector(X, 6);
     RedimensionStatVector(Y, X.Dimension);

     X.Vector^[1]:= 0.1;
     X.Vector^[2]:= 0.4;
     X.Vector^[3]:= 0.5;
     X.Vector^[4]:= 0.7;
     X.Vector^[5]:= 0.7;
     X.Vector^[6]:= 0.9;

     Y.Vector^[1]:= 0.61;
     Y.Vector^[2]:= 0.92;
     Y.Vector^[3]:= 0.99;
     Y.Vector^[4]:= 1.52;
     Y.Vector^[5]:= 1.47;
     Y.Vector^[6]:= 2.03;

     { Populate statmatrix M.
       Recall from least squares that the structure, for this function, is:
       [[ n              sum(xi)           sum(sin(xi))           sum(exp(xi))
          sum(xi)        sum(xi*xi)        sum(xi*sin(xi))        sum(xi*exp(xi))
          sum(sin(xi))   sum(xi*sin(xi))   sum(sqr(sin(xi)))      sum(sin(xi)*exp(xi))
          sum(exp(xi))   sum(xi*exp(xi))   sum(sin(xi)*exp(xi))   sum(sqr(exp(xi)))     ]]
       with xi the i-th component of statvector X.
       We will find the matrix elements by using the mathematical function parser of
       ChroMathsome - Statistics.                                                      }
     RedimensionStatmatrix(M, 4, 4);
     M.Matrix^[1]^[1]:= X.Dimension; // # of data points
     M.Matrix^[1]^[2]:= Sum(X);
     M.Matrix^[2]^[1]:= Sum(X);
     M.Matrix^[2]^[2]:= SumOfSquares(X);
     CopyStatVector(X1, X); // Use X1 to apply functions on
     StatVectorIterateVector(X1, 'sin(x)');
     M.Matrix^[3]^[1]:= Sum(X1);
     CopyStatVector(X1, X);
     StatVectorIterateVector(X1, 'exp(x)');
     M.Matrix^[4]^[1]:= Sum(X1);
     CopyStatVector(X1, X);
     StatVectorIterateVector(X1, 'x*sin(x)');
     M.Matrix^[3]^[2]:= Sum(X1);
     CopyStatVector(X1, X);
     StatVectorIterateVector(X1, 'x*exp(x)');
     M.Matrix^[4]^[2]:= Sum(X1);
     M.Matrix^[1]^[3]:= M.Matrix^[3]^[1]; // M is a symmetric matrix!
     M.Matrix^[2]^[3]:= M.Matrix^[3]^[2];
     CopyStatVector(X1, X);
     StatVectorIterateVector(X1, 'sqr(sin(x))');
     M.Matrix^[3]^[3]:= Sum(X1);
     CopyStatVector(X1, X);
     StatVectorIterateVector(X1, 'sin(x)*exp(x)');
     M.Matrix^[4]^[3]:= Sum(X1);

     M.Matrix^[1]^[4]:= M.Matrix^[4]^[1];
     M.Matrix^[2]^[4]:= M.Matrix^[4]^[2];
     M.Matrix^[3]^[4]:= M.Matrix^[4]^[3];
     CopyStatVector(X1, X);
     StatVectorIterateVector(X1, 'sqr(exp(x))');
     M.Matrix^[4]^[4]:= Sum(X1);

     Memo1.Lines.Add('Statmatrix M: ');
     Write_StatMatrix(M);

     { Populate statvectors Y1 in M * A = Y1; A contains the coefficients we're looking for.  }
     { Again, recall from least squares the elements of Y for this function:
       [ sum(yi)
         sum(xi*yi)
         sum(sin(xi)*yi)
         sum(exp(xi)*yi) ]
       with yi, the i-th component of statvector Y, xi ... }
     det:= 0;
     RedimensionStatVector(Y1, 4);
     Y1.Vector^[1]:= Sum(Y);
     Y1.Vector^[2]:= SumOfPowerXPowerY(X, Y, 1, 1);
     CopyStatVector(X1, X);
     StatVectorIterateVector(X1, 'sin(x)');
     for i:= 1 to X1.Dimension do
     det:= det + X1.Vector^[i]*Y.Vector^[i];
     Y1.Vector^[3]:= det;
     CopyStatVector(X1, X);
     det:= 0;
     StatVectorIterateVector(X1, 'exp(x)');
     for i:= 1 to X1.Dimension do
     det:= det + X1.Vector^[i]*Y.Vector^[i];
     Y1.Vector^[4]:= det;

     Memo1.Lines.Add('Statvector Y1: ');
     Write_StatVector(Y1);

     { Now, all needed matrices and vectors are initialized.
       Next, we need to solve the system M * A = Y1 to find vector A containing the coefficients
       of f(x) = a1 + a2*x + a3*sin(x) + a4*exp(x).
       For the sake of simplicity, we'll use Cramer's rule to solve this system.                    }
     RedimensionStatVector(X1, 1);
     { 1. Calculate determinant of M.
          First, we'll make a copy of M since we need the original matrix later.  }
     CopyStatMatrix(M1, M);
     det:= StatMatrixDeterminant(M1);
     if det = 0 then ShowMessage('Can not solve system!') // or whatever
     else
     { 2. Calculate coefficients by solving the system.
          We will save the coefficients in a vector X1 for later use. }
     begin
         for i:= 1 to M.Row do
         begin
            CopyStatMatrix(M1, M);
            StatMatrixReplaceColumn(M1, Y1, i);
            cf:= StatMatrixDeterminant(M1) / det;
            Memo1.Lines.Add('Coefficient a' + IntToStr(i) + ': ' + FloatToStr(cf));
            if i = 1 then X1.Vector^[1]:= cf else
            StatVectorAddElement(X1, cf);
         end;
     end;

     { Finally - Test, let's find the deviation in every point to get an idea of the fit. }
     for i:= 1 to X.Dimension do
     begin
        det:= X1.Vector^[1] + X1.Vector^[2]*X.Vector^[i] + X1.Vector^[3]*sin(X.Vector^[i]) + X1.Vector^[4]*exp(X.Vector^[i]);
        cf:= abs(det) - abs(Y.Vector^[i]);
        Memo1.Lines.Add('Deviation in point ' + IntToStr(i) + ': ' + FloatToStr(cf));
     end;

     finally
     DestroyStatMatrix(M);
     DestroyStatMatrix(M1);
     DestroyStatVector(X);
     DestroyStatVector(X1);
     DestroyStatVector(Y);
     DestroyStatVector(Y1);
     end;
end;

end.
